//------------------------------------------------------------------------------
//	$Id: simpson.js,v 1.23 2018/10/03 17:35:22 wolf Exp $
//------------------------------------------------------------------------------
//	This program is free software. It comes without any warranty, to
//	the extent permitted by applicable law. You can redistribute it
//	and/or modify it under the terms of the Do What The Fuck You Want
//	To Public License, Version 2, as published by Sam Hocevar. See
//	http://www.wtfpl.net/ for more details.
//------------------------------------------------------------------------------

/*jsl:option explicit*/

var SIMPSON = {};

SIMPSON.getNumberSize = function (maxValue, nDigits) {

	var n = 1 + (nDigits? nDigits+1 : 0);
	while (maxValue >= 10) { n++; maxValue /= 10; }

	return n;
};

SIMPSON.ce = function (tagName, parent) {

	var element = document.createElement(tagName);
	if (parent) parent.appendChild(element);
	return element;
};

SIMPSON.ct = function (text) {

	return document.createTextNode(text);
};

SIMPSON.trim = function (s) {

	if (!s) return '';
	return s.replace(/^\s+/,'').replace(/\s+$/,'');
};

SIMPSON.getIntegerInput = function (input, minValue, maxValue) {

	var value = parseInt (this.trim(input.value), 10);

	if (isNaN(value) || value < minValue) value = minValue;
	if (value > maxValue) value = maxValue;

	input.value = value;

	return value;
};

SIMPSON.percent = function (quot) {

	if (isNaN (quot)) return '';

	var percent = 100 * (quot - 1);

	return (percent>0? '+' : '') + percent.toFixed(1) + '%';
};

SIMPSON.fixed = function (value, digits) {

	return isNaN(value)? '' : value.toFixed(digits);
};

SIMPSON.doCalc = function (ctrl) {

	var nGroups    = ctrl.groups.length;
	var nFunctions = ctrl.functions.length;
	var xnGroups   = ctrl.withRowSum? nGroups+1 : nGroups;

	//----------------------------------------------------------------------
	//	Per fuction
	//----------------------------------------------------------------------

	var allCnt = [];
	var allSum = [];

	for (var functionIndex in ctrl.functions) {

		//--------------------------------------------------------------
		//	compute sum and average (and colum and row sum)
		//--------------------------------------------------------------

		var input = ctrl.valueInputs[functionIndex];
		var value = this.getIntegerInput(input, ctrl.minValue, ctrl.maxValue);

		var saldoCnt = 0;
		var saldoSum = 0;

		for (var groupIndex=0; groupIndex<nGroups; groupIndex++) {

			input = ctrl.inputs[functionIndex][groupIndex].cnt;

			var cnt = this.getIntegerInput(input, ctrl.minCount, ctrl.maxCount);
			var sum = cnt * value;

			ctrl.inputs[functionIndex][groupIndex].sum.value = sum;
			ctrl.inputs[functionIndex][groupIndex].avg.value = this.fixed(sum/cnt, 2);

			//------------------------------------------------------
			//	row sum
			//------------------------------------------------------

			saldoCnt += cnt;
			saldoSum += sum;

			//------------------------------------------------------
			//	col sum
			//------------------------------------------------------

			allCnt[groupIndex] = (allCnt[groupIndex] || 0) + cnt;
			allSum[groupIndex] = (allSum[groupIndex] || 0) + sum;
		}

		//--------------------------------------------------------------
		//	set row sum
		//--------------------------------------------------------------

		if (ctrl.withRowSum) {

			ctrl.inputs[functionIndex][nGroups].cnt.value = saldoCnt;
			ctrl.inputs[functionIndex][nGroups].sum.value = saldoSum;
			ctrl.inputs[functionIndex][nGroups].avg.value = this.fixed(saldoSum/saldoCnt, 2);

			//------------------------------------------------------
			//	col sum
			//------------------------------------------------------

			allCnt[nGroups] = (allCnt[nGroups] || 0) + saldoCnt;
			allSum[nGroups] = (allSum[nGroups] || 0) + saldoSum;
		}
	}

	//----------------------------------------------------------------------
	//	Col sums
	//----------------------------------------------------------------------

	var allAvg = [];

	for (groupIndex=0; groupIndex < xnGroups; groupIndex++) {

		saldoCnt = allCnt[groupIndex];
		saldoSum = allSum[groupIndex];

		var saldoAvg = saldoSum / saldoCnt;

		ctrl.inputs[nFunctions][groupIndex].cnt.value = saldoCnt;
		ctrl.inputs[nFunctions][groupIndex].sum.value = saldoSum;
		ctrl.inputs[nFunctions][groupIndex].avg.value = this.fixed(saldoAvg, 2);

		allAvg[groupIndex] = saldoAvg;
	}

	//----------------------------------------------------------------------
	//	Difference to average
	//----------------------------------------------------------------------

	if (ctrl.withRowSum) {

		for (groupIndex in ctrl.groups) {

			var delta = allAvg[groupIndex] / allAvg[nGroups ];
			ctrl.deltaInputs[groupIndex].value = SIMPSON.percent(delta);
		}
	}

	//----------------------------------------------------------------------
	//	Difference to left neighbour
	//----------------------------------------------------------------------

	for (groupIndex in ctrl.groups) {

		delta = allAvg[groupIndex] / allAvg[(groupIndex - 1 + nGroups) % nGroups];

		ctrl.compaInputs[groupIndex].value = SIMPSON.percent(delta);

		if (ctrl.smileIcons[groupIndex]) ctrl.smileIcons[groupIndex].style.display = delta>1? '' : 'none';
		if (ctrl.frownIcons[groupIndex]) ctrl.frownIcons[groupIndex].style.display = delta<1? '' : 'none';
	}
};

SIMPSON.create = function (id, config) {

	var container = document.getElementById(id);
	if (!container) return null;

	//----------------------------------------------------------------------
	//	text labels
	//----------------------------------------------------------------------

	var labelCnt          = config.labelCnt          || '#';
	var labelSum          = config.labelSum          || '\u2211';
	var labelAvg          = config.labelAvg          || '\u2205';
	var labelAllGroups    = config.labelAllGroups    || '\u2211';
	var labelAllFunctions = config.labelAllFunctions || '\u2211';

	//----------------------------------------------------------------------
	//	config values
	//----------------------------------------------------------------------

	var withRowSum = config.withRowSum|| false;

	var minCount   = config.minCount ||    0;
	var maxCount   = config.maxCount ||  100;

	var minValue   = config.minValue ||    1;
	var maxValue   = config.maxValue || 1000;

	var functions  = config.functions;
	var groups     = config.groups;

	var nGroups    = groups.length;
	var nFunctions = functions.length;
	var nElements  = 3; // count, sum, average
	var xnGroups   = withRowSum? nGroups+1 : nGroups;

	var valSize = this.getNumberSize (maxValue);
	var cntSize = this.getNumberSize (maxCount * nFunctions * nGroups);
	var sumSize = this.getNumberSize (maxCount * nFunctions * nGroups * maxValue);
	var avgSize = this.getNumberSize (maxValue, config.digits);
	var pctSize = 7; // +999.9%

	//----------------------------------------------------------------------
	//	main description (return result)
	//----------------------------------------------------------------------

	var ctrl = {

		withRowSum:  withRowSum,

		minCount:    minCount,
		maxCount:    maxCount,
		minValue:    minValue,
		maxValue:    maxValue,

		functions:   functions,
		groups:      groups,

		inputs:      [],
		valueInputs: [],
		deltaInputs: [],
		compaInputs: [],
		frownIcons:  [],
		smileIcons:  [],

		setValues: function (newValue) {

			var values = newValue instanceof Array? newValue.slice() : [newValue];

			for (var valueIndex in this.valueInputs) {

				var value = parseInt (values[0], 10);

				if (isNaN(value) || value < this.minValue) value = this.minValue;
				if (value > this.maxValue) value = this.maxValue;

				this.valueInputs[valueIndex].value = value;

				if (values.length>=2) values.shift();
			}

			SIMPSON.doCalc(this);
		},

		setCounts: function (functionIndex, newCount) {

			var inputs = this.inputs[functionIndex];
			var counts = newCount instanceof Array? newCount.slice() : [newCount];

			for (var groupIndex in this.groups) {

				var count = parseInt (counts[0], 10);

				if (isNaN(count) || count < this.minCount) count = this.minCount;
				if (count > this.maxCount) count = this.maxCount;

				inputs[groupIndex].cnt.value = count;

				if (counts.length>=2) counts.shift();
			}

			SIMPSON.doCalc(this);
		}
	};

	for (var functionIndex=0; functionIndex<=nFunctions; functionIndex++) {

		ctrl.inputs[functionIndex] = [];
		for (var groupIndex=0; groupIndex<=nGroups; groupIndex++) {

			ctrl.inputs[functionIndex][groupIndex] = {};
		}
	}

	//----------------------------------------------------------------------
	//	<table>
	//----------------------------------------------------------------------

	var table = this.ce('table', container);
	table.className = 'simpson';

	if (config.tableName) {

		var caption = this.ce('caption', table);
		caption.appendChild(this.ct(config.tableName));
	}

	var thead, tfoot, tbody, tr, th, td, input, img;

	//----------------------------------------------------------------------
	//	<colgroup>
	//----------------------------------------------------------------------

	var colgroup = this.ce('colgroup', table);
	var col;

	if (config.groupFirst) {

		col = this.ce('col', colgroup);
		col.setAttribute ('span', 2);

		for (groupIndex in groups) {

			col = this.ce('col', colgroup);
			col.setAttribute ('span', nElements);
			col.style = config.groups[groupIndex].style || '';
		}

		if (withRowSum) {

			col = this.ce('col', colgroup);
			col.setAttribute ('span', nElements);
		}

	} else {

		col = this.ce('col', colgroup);
		col.setAttribute ('span', 2);

		for (var e=0; e<nElements; e++) {

			for (groupIndex in groups) {

				col = this.ce('col', colgroup);
				col.style = config.groups[groupIndex].style || '';
			}

			if (withRowSum) this.ce('col', colgroup);
		}
	}

	//----------------------------------------------------------------------
	//	<thead><tr> (1)
	//----------------------------------------------------------------------

	thead = this.ce('thead', table);

	tr = this.ce('tr', thead);

	th = this.ce('th', tr);
	th.setAttribute ('rowspan', 2);
	th.setAttribute ('scope', 'col');
	th.appendChild(this.ct(config.functionName));

	th = this.ce('th', tr);
	th.setAttribute ('rowspan', 2);
	th.setAttribute ('scope', 'col');
	th.appendChild(this.ct(config.valueName));

	if (config.groupFirst) {

		for (groupIndex in groups) {

			var name = groups[groupIndex].name;

			th = this.ce('th', tr);
			th.setAttribute ('colspan', nElements);
			th.setAttribute ('scope', 'col');
			th.appendChild(this.ct(name));
		}

		if (withRowSum) {

			th = this.ce('th', tr);
			th.setAttribute ('colspan', nElements);
			th.setAttribute ('scope', 'col');
			th.appendChild(this.ct(labelAllGroups));
		}

	} else {

		th = this.ce('th', tr);
		th.setAttribute ('colspan', xnGroups);
		th.setAttribute ('scope', 'col');
		th.setAttribute ('class', 'nogroup');
		th.appendChild(this.ct(labelCnt));

		th = this.ce('th', tr);
		th.setAttribute ('colspan', xnGroups);
		th.setAttribute ('scope', 'col');
		th.setAttribute ('class', 'nogroup');
		th.appendChild(this.ct(labelSum));

		th = this.ce('th', tr);
		th.setAttribute ('colspan', xnGroups);
		th.setAttribute ('scope', 'col');
		th.setAttribute ('class', 'nogroup');
		th.appendChild(this.ct(labelAvg));
		tr.appendChild (th);
	}

	//----------------------------------------------------------------------
	//	Table header (2)
	//----------------------------------------------------------------------

	tr = this.ce('tr', thead);

	if (config.groupFirst) {

		for (groupIndex=0; groupIndex<xnGroups; groupIndex++) {

			th = this.ce('th', tr);
			th.setAttribute ('scope', 'col');
			th.appendChild(this.ct(labelCnt));

			th = this.ce('th', tr);
			th.setAttribute ('scope', 'col');
			th.appendChild(this.ct(labelSum));

			th = this.ce('th', tr);
			th.setAttribute ('scope', 'col');
			th.appendChild(this.ct(labelAvg));
			tr.appendChild (th);
		}

	} else {

		for (var i=0; i<nElements; i++) { // count, sum, avg

			for (groupIndex=0; groupIndex<nGroups; groupIndex++) {

				th = this.ce('th', tr);
				th.setAttribute ('scope', 'col');
				th.appendChild(this.ct(groups[groupIndex].name));
			}

			if (withRowSum) {

				th = this.ce('th', tr);
				th.setAttribute ('scope', 'col');
				th.appendChild(this.ct(labelAllGroups));
			}
		}
	}

	//----------------------------------------------------------------------
	//	Table footer
	//----------------------------------------------------------------------

	tfoot = this.ce('tfoot', table);

	//----------------------------------------------------------------------
	//	Table footer (line 1)
	//----------------------------------------------------------------------

	if (withRowSum) {
		tr = this.ce('tr', tfoot);
		tr.setAttribute('class', 'delta');

		if (config.groupFirst) {

			//------------------------------------------------------
			//	filler
			//------------------------------------------------------

			th = this.ce('th', tr);
			th.setAttribute ('colspan', 2);
			th.setAttribute ('scope', 'row');
			th.appendChild(this.ct('Differenz zum Durchschnitt:'));

			//--------------------------------------------------------------
			//	percent
			//--------------------------------------------------------------

			for (groupIndex in groups) {

				td = this.ce('td', tr);
				td.setAttribute ('colspan', nElements);

				input = this.ce ('input', td);
				input.setAttribute('disabled', true);
				input.setAttribute('size', pctSize);
				input.setAttribute('maxlength', pctSize);

				ctrl.deltaInputs[groupIndex] = input;
			}

			//------------------------------------------------------
			//	filler
			//------------------------------------------------------

			if (withRowSum) {

				td = this.ce('td', tr);
				td.setAttribute ('colspan', nElements);
				td.setAttribute ('rowspan', 2);
			}

		} else {

			//------------------------------------------------------
			//	filler
			//------------------------------------------------------

			th = this.ce('th', tr);
			th.setAttribute ('colspan', 2 + 2 * xnGroups);
			th.setAttribute ('scope', 'row');
			th.appendChild(this.ct('Differenz zum Durchschnitt:'));

			//--------------------------------------------------------------
			//	percent
			//--------------------------------------------------------------

			for (groupIndex in groups) {

				td = this.ce('td', tr);

				input = this.ce ('input', td);
				input.setAttribute('disabled', true);
				input.setAttribute('size', pctSize);
				input.setAttribute('maxlength', pctSize);

				ctrl.deltaInputs[groupIndex] = input;
			}

			//------------------------------------------------------
			//	filler
			//------------------------------------------------------

			if (withRowSum) this.ce('td', tr);
		}
	}

	//----------------------------------------------------------------------
	//	Table footer (Prozent 2)
	//----------------------------------------------------------------------

	tr = this.ce('tr', tfoot);
	tr.setAttribute('class', 'compa');

	if (config.groupFirst) {

		//------------------------------------------------------
		//	filler
		//------------------------------------------------------

		th = this.ce('th', tr);
		th.setAttribute ('colspan', 2);
		th.setAttribute ('scope', 'row');
		th.appendChild(this.ct('Differenz zur Nachbargruppe:'));

		//--------------------------------------------------------------
		//	percent
		//--------------------------------------------------------------

		for (groupIndex in groups) {

			var group = groups[groupIndex];

			td = this.ce('td', tr);
			td.setAttribute ('colspan', nElements);

			if (group.smileIcon) {

				img = this.ce('img', td);
				img.setAttribute('src', group.smileIcon);
				img.style.display = 'none';

				ctrl.smileIcons[groupIndex] = img;
			}

			if (group.frownIcon) {

				img = this.ce('img', td);
				img.setAttribute('src', group.frownIcon);
				img.style.display = 'none';

				ctrl.frownIcons[groupIndex] = img;
			}

			input = this.ce ('input', td);
			input.setAttribute('disabled', true);
			input.setAttribute('size', pctSize);
			input.setAttribute('maxlength', pctSize);

			ctrl.compaInputs[groupIndex] = input;
		}

	} else {

		//------------------------------------------------------
		//	filler
		//------------------------------------------------------

		th = this.ce('th', tr);
		th.setAttribute ('colspan', 2 + 2 * xnGroups);
		th.setAttribute ('scope', 'row');
		th.appendChild(this.ct('Differenz zur Nachbargruppe:'));

		//--------------------------------------------------------------
		//	percent
		//--------------------------------------------------------------

		for (groupIndex in groups) {

			group = groups[groupIndex];

			td = this.ce('td', tr);

			if (group.smileIcon) {

				img = this.ce('img', td);
				img.setAttribute('src', group.smileIcon);
				img.style.display = 'none';

				ctrl.smileIcons[groupIndex] = img;
			}

			if (group.frownIcon) {

				img = this.ce('img', td);
				img.setAttribute('src', group.frownIcon);
				img.style.display = 'none';

				ctrl.frownIcons[groupIndex] = img;
			}

			input = this.ce ('input', td);
			input.setAttribute('disabled', true);
			input.setAttribute('size', pctSize);
			input.setAttribute('maxlength', pctSize);

			ctrl.compaInputs[groupIndex] = input;
		}

		//------------------------------------------------------
		//	filler
		//------------------------------------------------------

		if (withRowSum) this.ce('td', tr);
	}

	//----------------------------------------------------------------------
	//	Table body (Funktionen)
	//----------------------------------------------------------------------

	tbody = this.ce('tbody', table);

	for (functionIndex in functions) {

		var funct = functions[functionIndex];

		tr = this.ce('tr', tbody);

		//--------------------------------------------------------------
		//	name + value
		//--------------------------------------------------------------

		th = this.ce('th', tr);
		th.setAttribute ('scope', 'row');
		th.appendChild(this.ct(funct.name + ':'));

		td = this.ce('td', tr);

		input = this.ce ('input', td);
		input.setAttribute('class', 'value');
		input.setAttribute('maxlength', valSize);
		input.setAttribute('size', valSize);
		input.setAttribute('value', funct.value);

		ctrl.valueInputs[functionIndex] = input;

		input.onblur  = function() { SIMPSON.doCalc(ctrl); };

		//--------------------------------------------------------------
		//	counts, sums, averages
		//--------------------------------------------------------------

		var counts = funct.count instanceof Array? funct.count.slice() : [funct.count];

		if (config.groupFirst) {

			for (groupIndex=0; groupIndex<xnGroups; groupIndex++) {

				//----------------------------------------------
				//	count
				//----------------------------------------------

				td = this.ce('td', tr);
				td.setAttribute('class', 'cnt');

				input = this.ce ('input', td);
				input.setAttribute('size', cntSize);
				input.setAttribute('maxlength', cntSize);

				ctrl.inputs[functionIndex][groupIndex].cnt = input;

				if (groupIndex < nGroups) {

					input.setAttribute('value', counts[0]);
					input.onblur  = function() { SIMPSON.doCalc(ctrl); };

					if (counts.length>=2) counts.shift();
				} else {
					input.setAttribute('disabled', true);
				}

				//----------------------------------------------
				//	sum
				//----------------------------------------------

				td = this.ce('td', tr);
				td.setAttribute('class', 'sum');

				input = this.ce ('input', td);
				input.setAttribute('disabled', true);
				input.setAttribute('size', sumSize);
				input.setAttribute('maxlength', sumSize);

				ctrl.inputs[functionIndex][groupIndex].sum = input;

				//----------------------------------------------
				//	avg
				//----------------------------------------------

				td = this.ce('td', tr);
				td.setAttribute('class', 'avg');

				input = this.ce ('input', td);
				input.setAttribute('disabled', true);
				input.setAttribute('size', avgSize);
				input.setAttribute('maxlength', avgSize);

				ctrl.inputs[functionIndex][groupIndex].avg = input;
			}

		} else {

			//------------------------------------------------------
			//	counts
			//------------------------------------------------------

			for (groupIndex=0; groupIndex<xnGroups; groupIndex++) {

				td = this.ce('td', tr);
				td.setAttribute('class', 'cnt');

				input = this.ce ('input', td);
				input.setAttribute('size', cntSize);
				input.setAttribute('maxlength', cntSize);

				if (groupIndex<nGroups) {

					input.setAttribute('value', counts[0]);
					input.onblur  = function() { SIMPSON.doCalc(ctrl); };

					if (counts.length>=2) counts.shift();
				} else {
					input.setAttribute('disabled', true);
				}

				ctrl.inputs[functionIndex][groupIndex].cnt = input;
			}

			//------------------------------------------------------
			//	sums
			//------------------------------------------------------

			for (groupIndex=0; groupIndex<xnGroups; groupIndex++) {

				td = this.ce('td', tr);
				td.setAttribute('class', 'sum');

				input = this.ce ('input', td);
				input.setAttribute('disabled', true);
				input.setAttribute('size', sumSize);
				input.setAttribute('maxlength', sumSize);

				ctrl.inputs[functionIndex][groupIndex].sum = input;
			}

			//------------------------------------------------------
			//	avgs
			//------------------------------------------------------

			for (groupIndex=0; groupIndex<xnGroups; groupIndex++) {

				td = this.ce('td', tr);
				td.setAttribute('class', 'avg');

				input = this.ce ('input', td);
				input.setAttribute('disabled', true);
				input.setAttribute('size', avgSize);
				input.setAttribute('maxlength', avgSize);

				ctrl.inputs[functionIndex][groupIndex].avg = input;
			}
		}
	}

	//----------------------------------------------------------------------
	//	Table body (Summe)
	//----------------------------------------------------------------------

	tbody = this.ce('tbody', table);

	tr = this.ce('tr', tbody);

	//--------------------------------------------------------------
	//	name + value
	//--------------------------------------------------------------

	th = this.ce('th', tr);
	tr.setAttribute('class', 'all');
	th.setAttribute('colspan', 2);
	th.setAttribute('scope', 'row');
	th.appendChild(this.ct(labelAllFunctions + ':'));

	if (config.groupFirst) {

		for (groupIndex=0; groupIndex<xnGroups; groupIndex++) {

			//----------------------------------------------
			//	count
			//----------------------------------------------

			td = this.ce('td', tr);
			td.setAttribute('class', 'cnt');

			input = this.ce ('input', td);
			input.setAttribute('disabled', true);
			input.setAttribute('size', cntSize);
			input.setAttribute('maxlength', cntSize);

			ctrl.inputs[nFunctions][groupIndex].cnt = input;

			//----------------------------------------------
			//	sum
			//----------------------------------------------

			td = this.ce('td', tr);
			td.setAttribute('class', 'sum');

			input = this.ce ('input', td);
			input.setAttribute('disabled', true);
			input.setAttribute('size', sumSize);
			input.setAttribute('maxlength', sumSize);

			ctrl.inputs[nFunctions][groupIndex].sum = input;

			//----------------------------------------------
			//	avgs
			//----------------------------------------------

			td = this.ce('td', tr);
			td.setAttribute('class', 'avg');

			input = this.ce ('input', td);
			input.setAttribute('disabled', true);
			input.setAttribute('size', avgSize);
			input.setAttribute('maxlength', avgSize);

			ctrl.inputs[nFunctions][groupIndex].avg = input;
		}
	} else {

		//------------------------------------------------------
		//	counts
		//------------------------------------------------------

		for (groupIndex=0; groupIndex<xnGroups; groupIndex++) {

			td = this.ce('td', tr);
			td.setAttribute('class', 'cnt');

			input = this.ce ('input', td);
			input.setAttribute('disabled', true);
			input.setAttribute('size', cntSize);
			input.setAttribute('maxlength', cntSize);

			ctrl.inputs[nFunctions][groupIndex].cnt = input;
		}

		//------------------------------------------------------
		//	sums
		//------------------------------------------------------

		for (groupIndex=0; groupIndex<xnGroups; groupIndex++) {

			td = this.ce('td', tr);
			td.setAttribute('class', 'sum');

			input = this.ce ('input', td);
			input.setAttribute('disabled', true);
			input.setAttribute('size', sumSize);
			input.setAttribute('maxlength', sumSize);

			ctrl.inputs[nFunctions][groupIndex].sum = input;
		}

		//------------------------------------------------------
		//	avgs
		//------------------------------------------------------

		for (groupIndex=0; groupIndex<xnGroups; groupIndex++) {

			td = this.ce('td', tr);
			td.setAttribute('class', 'avg');

			input = this.ce ('input', td);
			input.setAttribute('disabled', true);
			input.setAttribute('size', avgSize);
			input.setAttribute('maxlength', avgSize);

			ctrl.inputs[nFunctions][groupIndex].avg = input;
		}
	}

	this.doCalc (ctrl);

	return ctrl;
};

//------------------------------------------------------------------------------
//	$Id: simpson.js,v 1.23 2018/10/03 17:35:22 wolf Exp $
//------------------------------------------------------------------------------
